# -*- coding: utf-8 -*-

project = 'wwt_data_formats'
author = 'The AAS WorldWide Telescope Team'
copyright = '2019-2020 the .NET Foundation'

release = '0.4.4'  # cranko project-version
version = '.'.join(release.split('.')[:2])

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx_automodapi.automodapi',
    'sphinx_automodapi.smart_resolver',
    'numpydoc',
]

templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
language = None
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
pygments_style = 'sphinx'
todo_include_todos = False

html_theme = 'bootstrap-astropy'
html_theme_options = {
    'logotext1': 'wwt_data_formats',
    'logotext2': '',
    'logotext3': ':docs',
    'astropy_project_menubar': False,
}
html_static_path = ['_static']
htmlhelp_basename = 'wwtdataformatsdoc'

intersphinx_mapping = {
    'python': (
        'https://docs.python.org/3/',
        (None, 'http://data.astropy.org/intersphinx/python3.inv')
    ),

    'numpy': (
        'https://docs.scipy.org/doc/numpy/',
        (None, 'http://data.astropy.org/intersphinx/numpy.inv')
    ),

    'astropy': (
        'http://docs.astropy.org/en/stable/',
        None
    ),

    'requests': (
        'https://requests.readthedocs.io/en/stable/',
        None
    ),

    'traitlets': (
        'https://traitlets.readthedocs.io/en/stable/',
        None
    ),
}

numpydoc_show_class_members = False

nitpicky = True
nitpick_ignore = [
    ('py:class', 'traitlets.traitlets.MetaHasTraits'),
]

default_role = 'obj'

html_logo = 'images/logo.png'

linkcheck_retries = 5
linkcheck_timeout = 10



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the .NET Foundation
# Licensed under the MIT License.

"""Various enumerations

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
Bandpass
Classification
Constellation
DataSetType
FolderType
ProjectionType
SerEnum
'''.split()

from enum import Enum


class SerEnum(Enum):
    """A helper for enumerations that are serialized and deserialized from text.
    In some WWT cases, multiple textualizations have historically been used
    for the same value.

    """
    @classmethod
    def from_text(cls, text):
        """The default implementation here just looks up the appropriate enum instance
        from its value.

        """
        return cls(text)

class Bandpass(SerEnum):
    GAMMA = 'Gamma'
    XRAY = 'XRay'
    ULTRAVIOLET = 'Ultraviolet'
    VISIBLE = 'Visible'
    HYDROGEN_ALPHA = 'HydrogenAlpha'
    INFRARED = 'IR'
    MICROWAVE = 'Microwave'
    RADIO = 'Radio'
    VISIBLE_NIGHT = 'VisibleNight'


class Classification(SerEnum):
    UNSPECIFIED = ''
    STAR = 'Star'
    SUPERNOVA = 'Supernova'
    BLACKHOLE = 'BlackHole'
    NEUTRON_STAR = 'NeutronStar'
    DOUBLE_STAR = 'DoubleStar'
    MULTIPLE_STARS = 'MultipleStars'
    ASTERISM = 'Asterism'
    CONSTELLATION = 'Constellation'
    OPEN_CLUSTER = 'OpenCluster'
    GLOBULAR_CLUSTER = 'GlobularCluster'
    NEBULOUS_CLUSTER = 'NebulousCluster'
    NEBULA = 'Nebula'
    EMISSION_NEBULA = 'EmissionNebula'
    PLANETARY_NEBULA = 'PlanetaryNebula'
    REFLECTION_NEBULA = 'ReflectionNebula'
    DARK_NEBULA = 'DarkNebula'
    GIANT_MOLECULAR_CLOUD = 'GiantMolecularCloud'
    SUPERNOVA_REMNANT = 'SupernovaRemnant'
    INTERSTELLAR_DUST = 'InterstellarDust'
    QUASAR = 'Quasar'
    GALAXY = 'Galaxy'
    SPIRAL_GALAXY = 'SpiralGalaxy'
    IRREGULAR_GALAXY = 'IrregularGalaxy'
    ELLIPTICAL_GALAXY = 'EllipticalGalaxy'
    KNOT = 'Knot'
    PLATE_DEFECT = 'PlateDefect'
    CLUSTER_OF_GALAXIES = 'ClusterOfGalaxies'
    OTHER_NGC = 'OtherNGC'
    UNIDENTIFIED = 'Unidentified'
    SOLARS_YSTEM = 'SolarSystem'
    UNFILTERED = 'Unfiltered'
    STELLAR = 'Stellar'
    STELLAR_GROUPINGS = 'StellarGroupings'
    NEBULAE = 'Nebulae'
    GALACTIC = 'Galactic'
    OTHER = 'Other'

class Constellation(SerEnum):
    UNSPECIFIED = ""
    ANDROMEDA = "AND"
    ANTLIA = "ANT"
    APUS = "APS"
    AQUARIUS = "AQR"
    AQUILA = "AQL"
    ARA = "ARA"
    ARIES = "ARI"
    AURIGA = "AUR"
    BOOTES = "BOO"
    CAELUM = "CAE"
    CAMELOPARDALIS = "CAM"
    CANCER = "CNC"
    CANES_VENATICI = "CVN"
    CANIS_MAJOR = "CMA"
    CANIS_MINOR = "CMI"
    CAPRICORNUS = "CAP"
    CARINA = "CAR"
    CASSIOPEIA = "CAS"
    CENTAURUS = "CEN"
    CEPHEUS = "CEP"
    CETUS = "CET"
    CHAMAELEON = "CHA"
    CIRCINUS = "CIR"
    COLUMBA = "COL"
    COMA_BERENICES = "COM"
    CORONA_AUSTRALIS = "CRA"
    CORONA_BOREALIS = "CRB"
    CORVUS = "CRV"
    CRATER = "CRT"
    CRUX = "CRU"
    CYGNUS = "CYG"
    DELPHINUS = "DEL"
    DORADO = "DOR"
    DRACO = "DRA"
    EQUULEUS = "EQU"
    ERIDANUS = "ERI"
    FORNAX = "FOR"
    GEMINI = "GEM"
    GRUS = "GRU"
    HERCULES = "HER"
    HOROLOGIUM = "HOR"
    HYDRA = "HYA"
    HYDRUS = "HYI"
    INDUS = "IND"
    LACERTA = "LAC"
    LEO = "LEO"
    LEO_MINOR = "LMI"
    LEPUS = "LEP"
    LIBRA = "LIB"
    LUPUS = "LUP"
    LYNX = "LYN"
    LYRA = "LYR"
    MENSA = "MEN"
    MICROSCOPIUM = "MIC"
    MONOCEROS = "MON"
    MUSCA = "MUS"
    NORMA = "NOR"
    OCTANS = "OCT"
    OPHIUCHUS = "OPH"
    ORION = "ORI"
    PAVO = "PAV"
    PEGASUS = "PEG"
    PERSEUS = "PER"
    PHOENIX = "PHE"
    PICTOR = "PIC"
    PISCES = "PSC"
    PISCIS_AUSTRINUS = "PSA"
    PUPPIS = "PUP"
    PYXIS = "PYX"
    RETICULUM = "RET"
    SAGITTA = "SGE"
    SAGITTARIUS = "SGR"
    SCORPIUS = "SCO"
    SCULPTOR = "SCL"
    SCUTUM = "SCT"
    SERPENS_CAPUT = "SER1"
    SERPENS_CAUDA = "SER2"
    SEXTANS = "SEX"
    TAURUS = "TAU"
    TELESCOPIUM = "TEL"
    TRIANGULUM = "TRI"
    TRIANGULUM_AUSTRALE = "TRA"
    TUCANA = "TUC"
    URSA_MAJOR = "UMA"
    URSA_MINOR = "UMI"
    VELA = "VEL"
    VIRGO = "VIR"
    VOLANS = "VOL"
    VULPECULA = "VUL"


class DataSetType(SerEnum):
    EARTH = 'Earth'
    PLANET = 'Planet'
    SKY = 'Sky'
    PANORAMA = 'Panorama'
    SOLAR_SYSTEM = 'SolarSystem'
    SANDBOX = 'Sandbox'


class FolderType(SerEnum):
    UNSPECIFIED = ''
    EARTH = 'Earth'
    PLANET = 'Planet'
    SKY = 'Sky'
    PANORAMA = 'Panorama'


class ProjectionType(SerEnum):
    MERCATOR = 'Mercator'
    EQUIRECTANGULAR = 'Equirectangular'
    HEALPIX = 'Healpix'
    TAN = 'Tan'
    TOAST = 'Toast'
    SPHERICAL = 'Spherical'
    SKY_IMAGE = 'SkyImage'
    PLOTTED = 'Plotted'

    @classmethod
    def from_text(cls, text):
        if text == 'Tangent':
            return cls.TAN
        return cls(text)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the .NET Foundation
# Licensed under the MIT License.

"""
A basic HTTP server for local testing of WTML files and other WWT data products.

The key motivation for the existence of this server is that WWT WTML
"collection" files must contain absolute URLs. This means that if you're locally
testing a file, you need to include one set of (localhost) URLs, while the
production file must be different. Keeping these in sync is tedious and
error-prone.

This server is basically a generic local static-file server. But, if
specially-marked WTML files are requested, any relative URLs are rewritten
on-the-fly to be absolute relative to the server's address. That way, you can do
all of your development with the relative-URL files, and you only need to do one
substitution at the very end when the files are ready for upload to the
production server. (You can do this with ``wwtdatatool wtml rewrite-urls``.)

"""

from __future__ import absolute_import, division, print_function

__all__ = '''
run_server
'''.split()

from functools import partial
import http.server
import os.path

from .folder import Folder, make_absolutizing_url_mutator

try:
    from http.server import ThreadingHTTPServer as HTTPServerClass
except ImportError:
    from http.server import HTTPServer as HTTPServerClass


class WWTRequestHandler(http.server.SimpleHTTPRequestHandler):
    def end_headers(self):
        if self.command in ('GET', 'HEAD'):
            self.send_header('Access-Control-Allow-Headers', 'Content-Disposition,Content-Encoding,Content-Type')
            self.send_header('Access-Control-Allow-Origin', '*')
            self.send_header('Access-Control-Allow-Methods', 'GET,HEAD')
        return super(WWTRequestHandler, self).end_headers()

    def check_special_wtml(self):
        """
        Check whether this request is for a special synthetic WTML path, and if
        so, prep to respond to either a GET or a HEAD request. Returns a bytes
        of the response to write.

        If this request does not match a special synthetic WTML, returns None.
        """
        path = self.translate_path(self.path)

        if not path.endswith('.wtml'):
            return None

        local_wtml_path = path[:-5] + '_rel.wtml'
        if not os.path.exists(local_wtml_path):
            return None

        host = self.headers.get('Host')
        if host is None:
            host = f'{self.server.server_name}:{self.server.server_port}'
        baseurl = f'http://{host}{self.path}'
        self.log_message('special WTML: local %s, baseurl %s', local_wtml_path, baseurl)
        f = Folder.from_file(local_wtml_path)
        f.mutate_urls(make_absolutizing_url_mutator(baseurl))
        resp = f.to_xml_string().encode('utf-8')

        self.send_response(http.HTTPStatus.OK)
        self.send_header("Content-type", "application/x-wtml")
        self.send_header("Content-Length", str(len(resp)))
        self.end_headers()
        return resp

    def do_GET(self):
        maybe_content = self.check_special_wtml()
        if maybe_content is not None:
            self.wfile.write(maybe_content)
            return

        return super(WWTRequestHandler, self).do_GET()

    def do_HEAD(self):
        maybe_content = self.check_special_wtml()
        if maybe_content is not None:
            return

        return super(WWTRequestHandler, self).do_HEAD()


def run_server(settings):
    """
    Settings are defined in :func:`wwt_data_formats.cli.serve_getparser`.

    """
    server_address = ('', settings.port)
    handler_factory = partial(WWTRequestHandler, directory=settings.root_dir)

    with HTTPServerClass(server_address, handler_factory) as httpd:
        # Hack: the WWT JS engine special-cases 'localhost' and '127.0.0.1' so
        # that it doesn't start trying to proxy them if URLs result in 404s,
        # which is a common occurrence when working on tiled images. So ignore
        # the auto-detected server name and use one of those.
        server_name = '127.0.0.1'  # httpd.server_name

        print(f'listening at: http://{server_name}:{httpd.server_port}/')
        print()
        print('virtual root-directory WTML files with on-the-fly rewriting:')
        print()

        seen_any = False

        for bn in os.listdir(settings.root_dir):
            if bn.endswith('_rel.wtml'):
                virtual = bn[:-9] + '.wtml'
                print(f'    http://{server_name}:{httpd.server_port}/{virtual}')
                seen_any = True

        if not seen_any:
            print('    (none)')

        print()

        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            print()
            print('(interrupted)')




# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the .NET Foundation
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

from argparse import Namespace
from http.server import HTTPServer
import pytest
import requests
from threading import Thread
from time import sleep

from . import test_path
from .. import folder, server

TIMEOUT = 120

class TestServer(object):
    @classmethod
    def setup_class(cls):
        cls.server_port = 33007 # just hope it's not taken

        settings = Namespace()
        settings.port = cls.server_port
        settings.root_dir = test_path()

        # Note: there can be a race condition between server startup and the
        # execution of the tests. There doesn't seem to be any mechanism to
        # enable the tests to rigorously wait for the server to be fully started
        # up.
        cls.thread = Thread(target=lambda: server.run_server(settings))
        cls.thread.setDaemon(True)
        cls.thread.start()

        # In an attempt to help mitigate the above issue, set up requests to
        # retry requests if they fail. This is hardly expected to help since if
        # the server isn't yet running, the failures and retries will occur
        # instantly, but it's not bad to have an example of the pattern anyway.
        from requests.adapters import HTTPAdapter
        from requests.packages.urllib3.util.retry import Retry
        retry_strategy = Retry(
            total=10,
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        cls.session = requests.Session()
        cls.session.mount("https://", adapter)
        cls.session.mount("http://", adapter)

    def test_wtml_rewrite(self):
        base_url = 'http://localhost:{}/'.format(self.server_port)

        # More race condition mitigation
        for attempt in range(10):
            try:
                f = folder.Folder.from_url(base_url + 'test1.wtml', timeout=TIMEOUT, session=self.session)
                break
            except requests.ConnectionError:
                sleep(5)

        assert f.children[0].thumbnail == base_url + 'thumb.jpg'


def test_smoke():
    """Dumb smoketest."""

    server_address = ('', 0)

    with HTTPServer(server_address, server.WWTRequestHandler) as httpd:
        pass



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the .NET Foundation
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import os.path
import shutil
import tempfile
import pytest

from .. import cli, filecabinet


@pytest.fixture
def tempdir():
    d = tempfile.mkdtemp()
    yield d
    shutil.rmtree(d)


def test_writer():
    with pytest.raises(ValueError):
        fcw = filecabinet.FileCabinetWriter()
        fcw.add_file_with_data('bad-file-not-bytes', u'This is not bytes.')


def test_cli(tempdir):
    "Simple smoke test to see if it runs at all."

    prev_dir = os.getcwd()
    os.chdir(tempdir)

    with open('file1.txt', 'wt') as f:
        print('Hello world', file=f)

    cli.entrypoint(['cabinet', 'pack', 'cabinet.wwtl', 'file1.txt'])

    os.remove('file1.txt')
    cli.entrypoint(['cabinet', 'unpack', 'cabinet.wwtl'])

    # Windows can't remove the temp tree unless we chdir out of it.
    os.chdir(prev_dir)


# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the .NET Foundation
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

from mock import Mock
import os.path
import shutil
import tempfile
import pytest
from xml.etree import ElementTree as etree

from . import assert_xml_trees_equal
from .. import cli, folder, imageset, place


@pytest.fixture
def tempdir():
    d = tempfile.mkdtemp()
    yield d
    shutil.rmtree(d)


BASIC_XML_STRING = '''
<Folder MSRCommunityId="0" MSRComponentId="0" Permission="0"
        Browseable="True" Group="Explorer" Searchable="True" Type="Sky" />
'''

ROOT_XML_STRING = '''
<Folder MSRCommunityId="0" MSRComponentId="0" Permission="0"
        Browseable="True" Group="Explorer" Searchable="True" Type="Sky">
    <Folder Url="http://example.com/child1.wtml" />
    <Place MSRCommunityId="0" MSRComponentId="0" Permission="0"
        Angle="0.0" AngularSize="0.0" DataSetType="Earth" Dec="0.0" Distance="0.0"
        DomeAlt="0.0" DomeAz="0.0" Lat="0.0" Lng="0.0" Magnitude="0.0"
        Opacity="100.0" RA="0.0" Rotation="0.0" ZoomLevel="0.0">
    </Place>
</Folder>
'''

CHILD1_XML_STRING = '''
<Folder Name="Child1" MSRCommunityId="0" MSRComponentId="0" Permission="0"
        Browseable="True" Group="Explorer" Searchable="True" Type="Sky">
    <Place MSRCommunityId="0" MSRComponentId="0" Permission="0"
        Angle="0.0" AngularSize="0.0" DataSetType="Earth" Dec="0.0" Distance="0.0"
        DomeAlt="0.0" DomeAz="0.0" Lat="0.0" Lng="0.0" Magnitude="0.0"
        Opacity="100.0" RA="0.0" Rotation="0.0" ZoomLevel="0.0">
    </Place>
</Folder>
'''

def fake_request_session_send(request, **kwargs):
    rv = Mock()

    if request.url == 'http://example.com/root.wtml':
        rv.text = ROOT_XML_STRING
    elif request.url == 'http://example.com/child1.wtml':
        rv.text = CHILD1_XML_STRING
    else:
        raise Exception(f'unexpected URL to fake requests.Session.send(): {request.url}')

    return rv


@pytest.fixture
def fake_requests(mocker):
    m = mocker.patch('requests.Session.send')
    m.side_effect = fake_request_session_send


def test_basic_xml():
    expected_xml = etree.fromstring(BASIC_XML_STRING)
    f = folder.Folder()
    observed_xml = f.to_xml()
    assert_xml_trees_equal(expected_xml, observed_xml)

    # At the moment we can't check that this object equals the one we
    # constructed programmatically, but at least we can test that the
    # deserialization code doesn't crash.
    f = folder.Folder.from_xml(expected_xml)


def test_children():
    expected_str = '''
<Folder Browseable="True" Group="Explorer" Searchable="True" Type="Sky"
        MSRCommunityId="0" MSRComponentId="0" Permission="0">
  <Place MSRCommunityId="0" MSRComponentId="0" Permission="0"
         Angle="0.0" AngularSize="0.0" DataSetType="Earth" Dec="0.0" Distance="0.0"
         DomeAlt="0.0" DomeAz="0.0" Lat="0.0" Lng="0.0" Magnitude="0.0"
         Opacity="100.0" RA="0.0" Rotation="0.0" ZoomLevel="0.0">
  </Place>
  <Place MSRCommunityId="0" MSRComponentId="0" Permission="0"
         Angle="0.0" AngularSize="0.0" DataSetType="Earth" Dec="0.0" Distance="0.0"
         DomeAlt="0.0" DomeAz="0.0" Lat="0.0" Lng="0.0" Magnitude="0.0"
         Opacity="100.0" RA="0.0" Rotation="0.0" ZoomLevel="0.0">
  </Place>
</Folder>
'''
    expected_xml = etree.fromstring(expected_str)
    f = folder.Folder()
    f.children.append(place.Place())
    f.children.append(place.Place())
    observed_xml = f.to_xml()
    assert_xml_trees_equal(expected_xml, observed_xml)

    f = folder.Folder.from_xml(expected_xml)


def test_walk():
    f0 = folder.Folder()
    f1 = folder.Folder()
    pl0 = place.Place()
    pl1 = place.Place()
    is0 = imageset.ImageSet()

    f0.children = [pl0, f1]
    f1.children = [is0, pl1]

    expected = [
        (0, (), f0),
        (1, (0, ), pl0),
        (1, (1, ), f1),
        (2, (1, 0), is0),
        (2, (1, 1), pl1),
    ]

    observed = list(f0.walk(download=False))
    assert observed == expected


def test_from_url():
    "Note that this test hits the network."
    url = 'http://www.worldwidetelescope.org/wwtweb/catalog.aspx?W=ExploreRoot'
    folder.Folder.from_url(url)  # just test that we don't crash


def test_fetch_tree(fake_requests, tempdir):
    "Simple smoke test to see whether it crashes."

    def on_fetch(url):
        pass

    folder.fetch_folder_tree('http://example.com/root.wtml', tempdir, on_fetch)

    for item in folder.walk_cached_folder_tree(tempdir):
        pass


def test_basic_url_mutation():
    f = folder.Folder()
    f.url = '../updir/somewhere.wtml'
    f.mutate_urls(folder.make_absolutizing_url_mutator('https://example.com/subdir/'))
    assert f.url == 'https://example.com/updir/somewhere.wtml'

    from ..place import Place
    from ..imageset import ImageSet

    imgset = ImageSet()
    imgset.url = 'image.jpg'
    p = Place()
    p.background_image_set = imgset
    f.children.append(p)
    f.mutate_urls(folder.make_absolutizing_url_mutator('https://example.com/subdir/'))

    assert f.url == 'https://example.com/updir/somewhere.wtml'
    assert imgset.url == 'https://example.com/subdir/image.jpg'


def test_wtml_rewrite_urls(tempdir):
    prev_dir = os.getcwd()
    os.chdir(tempdir)

    f = folder.Folder()
    f.url = '../updir/somewhere.wtml'

    with open('index_rel.wtml', 'wt') as f_out:
        f.write_xml(f_out)

    cli.entrypoint(['wtml', 'rewrite-urls', 'index_rel.wtml', 'https://example.com/subdir/', 'index.wtml'])

    f = folder.Folder.from_file('index.wtml')
    assert f.url == 'https://example.com/updir/somewhere.wtml'

    # Windows can't remove the temp tree unless we chdir out of it.
    os.chdir(prev_dir)


# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the .NET Foundation
# Licensed under the MIT License.

"""A place that a WWT user can visit.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
Place
'''.split()

from argparse import Namespace
from traitlets import Float, Instance, Int, Unicode, UseEnum

from . import LockedXmlTraits, XmlSer
from .abcs import UrlContainer
from .enums import Classification, Constellation, DataSetType
from .imageset import ImageSet

class Place(LockedXmlTraits, UrlContainer):
    """A place that can be visited."""

    data_set_type = UseEnum(
        DataSetType,
        default_value = DataSetType.EARTH
    ).tag(xml=XmlSer.attr('DataSetType'))
    name = Unicode('').tag(xml=XmlSer.attr('Name'))
    ra_hr = Float(0.0).tag(xml=XmlSer.attr('RA'))
    dec_deg = Float(0.0).tag(xml=XmlSer.attr('Dec'))
    latitude = Float(0.0).tag(xml=XmlSer.attr('Lat'))
    longitude = Float(0.0).tag(xml=XmlSer.attr('Lng'))
    constellation = UseEnum(
        Constellation,
        default_value = Constellation.UNSPECIFIED
    ).tag(xml=XmlSer.attr('Constellation'))
    classification = UseEnum(
        Classification,
        default_value = Classification.UNSPECIFIED
    ).tag(xml=XmlSer.attr('Classification'))
    magnitude = Float(0.0).tag(xml=XmlSer.attr('Magnitude'))
    distance = Float(0.0).tag(xml=XmlSer.attr('Distance'))
    angular_size = Float(0.0).tag(xml=XmlSer.attr('AngularSize'))
    zoom_level = Float(0.0).tag(xml=XmlSer.attr('ZoomLevel'))
    rotation_deg = Float(0.0).tag(xml=XmlSer.attr('Rotation'))
    angle = Float(0.0).tag(xml=XmlSer.attr('Angle'))
    opacity = Float(100.0).tag(xml=XmlSer.attr('Opacity'))
    dome_alt = Float(0.0).tag(xml=XmlSer.attr('DomeAlt'))
    dome_az = Float(0.0).tag(xml=XmlSer.attr('DomeAz'))
    background_image_set = Instance(ImageSet, allow_none=True).tag(xml=XmlSer.wrapped_inner('BackgroundImageSet'))
    foreground_image_set = Instance(ImageSet, allow_none=True).tag(xml=XmlSer.wrapped_inner('ForegroundImageSet'))
    image_set = Instance(ImageSet, allow_none=True).tag(xml=XmlSer.inner('ImageSet'))
    thumbnail = Unicode('').tag(xml=XmlSer.attr('Thumbnail'))
    msr_community_id = Int(0).tag(xml=XmlSer.attr('MSRCommunityId'))
    """The ID number of the WWT Community that this content came from."""

    msr_component_id = Int(0).tag(xml=XmlSer.attr('MSRComponentId'))
    """The ID number of this content item on the WWT Communities system."""

    permission = Int(0).tag(xml=XmlSer.attr('Permission'))
    "TBD."

    xmeta = Instance(
        Namespace,
        args = (),
        help = 'XML metadata - a namespace object for attaching arbitrary text to serialize',
    ).tag(xml=XmlSer.ns_to_attr('X'))

    def _tag_name(self):
        return 'Place'

    def mutate_urls(self, mutator):
        if self.thumbnail:
            self.thumbnail = mutator(self.thumbnail)

        if self.background_image_set:
            self.background_image_set.mutate_urls(mutator)

        if self.foreground_image_set:
            self.foreground_image_set.mutate_urls(mutator)

        if self.image_set:
            self.image_set.mutate_urls(mutator)

    def as_imageset(self):
        """Return an ImageSet for this place if one is defined.

        Returns
        -------
        Either :class:`wwt_data_formats.imageset.ImageSet` or None.

        Notes
        -----
        If the :attr:`foreground_image_set` of this :class:`Place` is not
        None, it is returned. Otherwise, if its :attr:`image_set` is not
        None, that is returned. Otherwise, None is returned.

        """
        if self.foreground_image_set is not None:
            return self.foreground_image_set
        return self.image_set



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the .NET Foundation
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import numpy.testing as nt
import pytest
from xml.etree import ElementTree as etree

from . import assert_xml_trees_equal
from .. import imageset, enums, stringify_xml_doc, write_xml_doc


def test_basic_xml():
    expected_str = '''
<ImageSet MSRCommunityId="0" MSRComponentId="0" Permission="0"
          BandPass="Gamma" BaseDegreesPerTile="0.1" BaseTileLevel="1"
          BottomsUp="True" CenterX="1.234" CenterY="-0.31415"
          DataSetType="Planet" ElevationModel="False" FileType=".PNG" Generic="False"
          MeanRadius="0.0" Name="Test name"
          OffsetX="100.1" OffsetY="100.2" Projection="SkyImage"
          Rotation="5.4321" Sparse="False" StockSet="False" TileLevels="4"
          Url="http://example.org/{0}" WidthFactor="2">
  <Credits>Escaping &amp; Entities</Credits>
  <CreditsUrl>https://example.org/credits</CreditsUrl>
  <Description>Escaping &lt;entities&gt;</Description>
  <ThumbnailUrl>https://example.org/thumbnail.jpg</ThumbnailUrl>
</ImageSet>
'''
    expected_xml = etree.fromstring(expected_str)

    imgset = imageset.ImageSet()
    imgset.data_set_type = enums.DataSetType.PLANET
    imgset.name = 'Test name'
    imgset.url = 'http://example.org/{0}'
    imgset.width_factor = 2
    imgset.base_tile_level = 1
    imgset.tile_levels = 4
    imgset.base_degrees_per_tile = 0.1
    imgset.file_type = '.PNG'
    imgset.bottoms_up = True
    imgset.projection = enums.ProjectionType.SKY_IMAGE
    imgset.center_x = 1.234
    imgset.center_y = -0.31415
    imgset.offset_x = 100.1
    imgset.offset_y = 100.2
    imgset.rotation_deg = 5.4321
    imgset.band_pass = enums.Bandpass.GAMMA
    imgset.sparse = False
    imgset.credits = 'Escaping & Entities'
    imgset.credits_url = 'https://example.org/credits'
    imgset.thumbnail_url = 'https://example.org/thumbnail.jpg'
    imgset.description = 'Escaping <entities>'

    observed_xml = imgset.to_xml()
    assert_xml_trees_equal(expected_xml, observed_xml)


def test_wcs_1():
    expected_str = '''
<ImageSet MSRCommunityId="0" MSRComponentId="0" Permission="0"
          BandPass="Visible" BaseDegreesPerTile="4.870732233333334e-05"
          BaseTileLevel="0" BottomsUp="False" CenterX="83.633083" CenterY="22.0145"
          DataSetType="Sky" ElevationModel="False" FileType=".png" Generic="False"
          MeanRadius="0.0" OffsetX="1502.8507831457316" OffsetY="1478.8005935660037"
          Projection="SkyImage" Rotation="-0.29036478519000003" Sparse="True"
          StockSet="False" TileLevels="0" WidthFactor="2">
</ImageSet>
'''
    expected_xml = etree.fromstring(expected_str)

    wcs_keywords = {
        'CTYPE1': 'RA---TAN',
        'CTYPE2': 'DEC--TAN',
        'CRVAL1': 83.633083,
        'CRVAL2': 22.0145,
        'PC1_1': 0.9999871586199364,
        'PC1_2': 0.005067799840785529,
        'PC2_1': -0.005067799840785529,
        'PC2_2': 0.9999871586199364,
        'CRPIX1': 1503.8507831457316,
        'CRPIX2': 1479.8005935660037,
        'CDELT1': -4.870732233333334e-05,
        'CDELT2': 4.870732233333334e-05,
    }

    imgset = imageset.ImageSet()
    imgset.set_position_from_wcs(wcs_keywords, 3000, 3000)

    observed_xml = imgset.to_xml()
    assert_xml_trees_equal(expected_xml, observed_xml)

    wcs_roundtrip = imgset.wcs_headers_from_position()

    for kw in wcs_roundtrip.keys():
        expected = wcs_keywords[kw]
        observed = wcs_roundtrip[kw]

        if kw in ('CTYPE1', 'CTYPE2'):
            assert expected == observed
        else:
            nt.assert_almost_equal(expected, observed)


def test_misc_ser():
    expected_str = '''
<ImageSet BandPass="Visible" BaseDegreesPerTile="0.0" BaseTileLevel="0"
          BottomsUp="False" CenterX="0.0" CenterY="0.0" DataSetType="Sky" ElevationModel="False"
          FileType=".png" Generic="False" MeanRadius="0.0" MSRCommunityId="0" MSRComponentId="0"
          OffsetX="0.0" OffsetY="0.0" Permission="0" Projection="SkyImage"
          Rotation="0.0" Sparse="True" StockSet="False" TileLevels="0"
          Url="http://example.com/unspecified" WidthFactor="2" />
'''
    expected_xml = etree.fromstring(expected_str)

    imgset = imageset.ImageSet()
    imgset.url = 'http://example.com/unspecified'

    observed_xml = imgset.to_xml()
    assert_xml_trees_equal(expected_xml, observed_xml)

    expected_text = stringify_xml_doc(expected_xml)
    observed_text = imgset.to_xml_string()
    assert observed_text == expected_text

    from io import StringIO, BytesIO

    expected_strio = StringIO()
    observed_strio = StringIO()
    write_xml_doc(expected_xml, dest_stream=expected_strio, indent=False)
    imgset.write_xml(observed_strio, indent=False)
    assert observed_strio.getvalue() == expected_strio.getvalue()

    expected_bio = BytesIO()
    observed_bio = BytesIO()
    write_xml_doc(expected_xml, dest_stream=expected_bio, dest_wants_bytes=True)
    imgset.write_xml(observed_bio, dest_wants_bytes=True)
    assert observed_bio.getvalue() == expected_bio.getvalue()



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the .NET Foundation
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import pytest

from . import assert_xml_trees_equal, test_path
from .. import layers


def test_read_imageset():
    lcr = layers.LayerContainerReader.from_file(test_path('imagesetlayer.wwtl'))
    assert len(lcr.layers) == 1
    assert lcr.layers[0].image_set.file_type == '.tif'

    imgdata = lcr.read_layer_file(lcr.layers[0], lcr.layers[0].extension)
    assert len(imgdata) == 13224

    lcr.close()



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the .NET Foundation
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

__all__ = '''
FileCabinetReader
FileCabinetWriter
'''.split()

from collections import namedtuple, OrderedDict
from xml.etree import ElementTree as etree

from . import stringify_xml_doc

ReaderFileInfo = namedtuple('ReaderFileInfo', 'name offset size')


class FileCabinetReader(object):
    """Reader for a simple container format for other files.

    Unlike most of the other WWT data formats implemented in this package, the
    file cabinet is not an XML serialization of a data structure. Instead,
    it's a container format for bundling files together.

    """
    _stream = None
    _files = None

    def __init__(self, stream):
        self._stream = stream
        self._files = {}

        # We assume that the stream is going to be accessed randomly. There
        # might be cases in which the access is sequential, in which case we
        # could avoid the seeks and operate on pipes and the like, but we'll
        # cross that bridge if/when we get there -- this format is simple.

        stream.seek(0)

        # It seems safest to figure out the bounds of the header by
        # searching for the text of the form `HeaderSize="0x000001b2"`.
        # TODO: explicit handling if the magic string isn't found or if the
        # header size is implausible.

        header = stream.read(256)
        idx = header.index(b'HeaderSize="')
        size_hex = header[idx+12:idx+22]
        header_size = int(size_hex, 16)

        # Now we can read the full header and parse it.

        stream.seek(0)
        header = stream.read(header_size)
        header_doc = etree.fromstring(header)

        for file in header_doc.find('Files').iterfind('File'):
            name = file.get('Name')
            rel_offset = file.get('Offset')
            size = file.get('Size')

            if name is None or rel_offset is None or size is None:
                raise Exception('incomplete File record in FileCabinet')

            try:
                rel_offset = int(rel_offset)
                size = int(size)
            except Exception as e:
                raise Exception('malformed Offset or Size in File record in FileCabinet')

            # TODO: handle names with bad characters, implausible offsets and sizes, etc.

            if name in self._files:
                raise Exception('duplicated File record "{}" in FileCabinet'.format(name))

            self._files[name] = ReaderFileInfo(name, header_size + rel_offset, size)


    def close(self):
        """Close the underlying stream, making this object essentially unusable."""
        self._stream.close()
        self._stream = None


    def filenames(self):
        """Return an iterable of the names of the files in this cabinet."""
        return self._files.keys()


    def read_file(self, filename):
        """Read the specified file into memory in its entirety and return its contents.

        Returns bytes.
        """
        if self._stream is None:
            raise Exception('cannot read file "{}" with a closed FileCabinetReader'.format(filename))

        info = self._files.get(filename)
        if info is None:
            raise Exception('no such file "{}" in FileCabinet'.format(filename))

        self._stream.seek(info.offset)
        return self._stream.read(info.size)


WriterFileInfo = namedtuple('WriterFileInfo', 'name size contents')


class FileCabinetWriter(object):
    """Writer for a simple container format for other files.

    One day, we should support the ability to stream data into a cabinet
    without having to grossly buffer everything in memory. But today is not
    that day.

    """
    _files = None

    def __init__(self):
        self._files = OrderedDict()


    def add_file_with_data(self, name, data):
        """Add a file whose contents are stored in an in-memory buffer.

        The *data* argument should be a bytes object.

        """
        if not isinstance(data, bytes):
            raise ValueError('the data argument must be an instance of the bytes type')

        if name in self._files:
            raise ValueError('a file named \"{}\" has already been added'.format(name))

        size = len(data)
        self._files[name] = WriterFileInfo(name, size, data)


    def filenames(self):
        """Return an iterable of the names of the files in this cabinet."""
        return self._files.keys()


    def emit(self, stream):
        """Write out the contents of this cabinet to the target stream.

        """
        # Create the header structure.

        SIZE_PLACEHOLDER = '0xZYXWVUTS'
        cabinet = etree.Element('FileCabinet')
        cabinet.set('HeaderSize', SIZE_PLACEHOLDER)

        files = etree.SubElement(cabinet, 'Files')
        offset = 0

        for info in self._files.values():
            f = etree.SubElement(files, 'File')
            f.set('Name', info.name)
            f.set('Size', str(info.size))
            f.set('Offset', str(offset))
            offset += info.size

        # Serialize and patch in the actual header size. With a
        # non-pathological XML serialization, the HeaderSize item will occur
        # within the first SIZE_REGION bytes while the first filename will
        # occur beyond the first SIZE_REGION bytes, meaning that we'll be
        # resistant if someone tries to break us by using a filename that
        # includes SIZE_PLACEHOLDER.

        SIZE_REGION = 90
        header = stringify_xml_doc(cabinet, indent=True)
        header = header.encode('utf-8')
        size_ascii = '0x{:08x}'.format(len(header)).encode('us-ascii')
        filled_size = header[:SIZE_REGION].replace(SIZE_PLACEHOLDER.encode('us-ascii'), size_ascii)
        header = filled_size + header[SIZE_REGION:]

        stream.write(header)

        # The rest is straightforward.

        for info in self._files.values():
            stream.write(info.contents)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the .NET Foundation
# Licensed under the MIT License.

"""An image, possibly tiled, for display in WWT.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
ImageSet
'''.split()

import math
from traitlets import Bool, Float, Int, Unicode, UseEnum
from xml.etree import ElementTree as etree

from . import LockedXmlTraits, XmlSer
from .abcs import UrlContainer
from .enums import Bandpass, DataSetType, ProjectionType


class ImageSet(LockedXmlTraits, UrlContainer):
    """A set of images."""

    data_set_type = UseEnum(DataSetType, default_value=DataSetType.SKY).tag(xml=XmlSer.attr('DataSetType'))
    """The renderer mode to which these data apply.

    Possible values are ``"Earth"``, ``"Planet"``, ``"Sky"``, ``"Panorama"``,
    ``"SolarSystem"``, and ``"Sandbox"``.

    """
    reference_frame = Unicode('').tag(xml=XmlSer.attr('ReferenceFrame'))
    """TBD."""

    name = Unicode('').tag(xml=XmlSer.attr('Name'))
    """A name used to refer to this imageset.

    Various parts of the WWT internals reference imagesets by this name, so
    it should be distinctive.

    """
    url = Unicode('').tag(xml=XmlSer.attr('Url'))
    """The URL of the image data.

    Either a URL or a URL template. TODO: details

    """
    dem_url = Unicode('').tag(xml=XmlSer.attr('DemUrl'))
    """The URL of the DEM data.

    Either a URL or a URL template. TODO: details

    """
    width_factor = Int(2).tag(xml=XmlSer.attr('WidthFactor'))
    """This is a legacy parameter. Leave it at 2."""

    base_tile_level = Int(0).tag(xml=XmlSer.attr('BaseTileLevel'))
    """The level of the highest (coarsest-resolution) tiling available.

    This should be zero except for special circumstances.

    """
    quad_tree_map = Unicode('').tag(xml=XmlSer.attr('QuadTreeMap'))
    """TBD."""

    tile_levels = Int(0).tag(xml=XmlSer.attr('TileLevels'))
    """The number of levels of tiling.

    Should be zero for untiled images. An image with ``tile_levels = 1`` has been
    broken into four tiles, each 256x256 pixels. For ``tile_levels = 2``, there are
    sixteen tiles, and the padded height of the tiled area is ``256 * 2**2 = 1024``
    pixels. Image with dimensions of 2048 pixels or smaller do not need to be tiled,
    so if this parameter is nonzero it will usually be 4 or larger.

    """
    base_degrees_per_tile = Float(0.0).tag(xml=XmlSer.attr('BaseDegreesPerTile'))
    """The angular scale of the image.

    For untiled images, should be the pixel scale: the number of degrees per
    pixel in the vertical direction. Non-square pixels are not supported.

    For tiled images, this is the height of the image with its dimensions
    padded out to the next largest power of 2 for tiling purposes. If a square
    image is 1200 pixels tall and has a height of 0.016 deg, the padded height
    would be 2048 pixels and this parameter should be set to
    0.016 * 2048 / 1200 = 0.0273.

    """
    file_type = Unicode('.png').tag(xml=XmlSer.attr('FileType'))
    """The extension of the image file(s) in this set, including a leading period.

    """
    bottoms_up = Bool(False).tag(xml=XmlSer.attr('BottomsUp'))
    """TBD."""

    projection = UseEnum(
        ProjectionType,
        default_value = ProjectionType.SKY_IMAGE
    ).tag(xml=XmlSer.attr('Projection'))
    """The type of projection used to place this image on the sky.

    For untiled images, this should be "SkyImage". For tiled images, it should
    be "Tan". The :meth:`set_position_from_wcs` method will set this value
    appropriately based on :attr:`tile_levels`.

    """
    center_x = Float(0.0).tag(xml=XmlSer.attr('CenterX'))
    """The horizontal location of the center of the image’s projection coordinate
    system.

    For sky images, this is a right ascension in degrees.

    """
    center_y = Float(0.0).tag(xml=XmlSer.attr('CenterY'))
    """The vertical location of the center of the image’s projection coordinate
    system.

    For sky images, this is a declination in degrees.

    """
    offset_x = Float(0.0).tag(xml=XmlSer.attr('OffsetX'))
    """The horizontal positioning of the image relative to its projection
    coordinate system.

    For untiled sky images, the image is by default positioned such that its
    lower left lands at the center of the projection coordinate system (namely,
    ``center_x`` and ``center_y``). The offset is measured in pixels and moves
    the image leftwards. Therefore, ``offset_x = image_width / 2`` places the
    center of the image at ``center_x``. This parameter is therefore analogous
    to the WCS keyword ``CRVAL1``.

    For tiled sky images, the offset is measured in *degrees*, and a value of
    zero means that the *center* of the image lands at the center of the
    projection coordinate system.

    """
    offset_y = Float(0.0).tag(xml=XmlSer.attr('OffsetY'))
    """The vertical positioning of the image relative to its projection
    coordinate system.

    For untiled sky images, the image is by default positioned such that its
    lower left lands at the center of the projection coordinate system (namely,
    ``center_x`` and ``center_y``). The offset is measured in pixels and moves
    the image downwards. Therefore, ``offset_y = image_height / 2`` places the
    center of the image at ``center_y``. This parameter is therefore analogous
    to the WCS keyword ``CRVAL2``.

    For tiled sky images, the offset is measured in *degrees*, and a value of
    zero means that the *center* of the image lands at the center of the
    projection coordinate system.

    """
    rotation_deg = Float(0.0).tag(xml=XmlSer.attr('Rotation'))
    """The rotation of image’s projection coordinate system, in degrees.

    For sky images, this is East from North, i.e. counterclockwise.

    """
    band_pass = UseEnum(
        Bandpass,
        default_value = Bandpass.VISIBLE
    ).tag(xml=XmlSer.attr('BandPass'))
    """The bandpass of the image data."""

    sparse = Bool(True).tag(xml=XmlSer.attr('Sparse'))
    """TBD."""

    elevation_model = Bool(False).tag(xml=XmlSer.attr('ElevationModel'))
    """TBD."""

    stock_set = Bool(False).tag(xml=XmlSer.attr('StockSet'))
    """TBD."""

    generic = Bool(False).tag(xml=XmlSer.attr('Generic'))
    """TBD."""

    mean_radius = Float(0.0).tag(xml=XmlSer.attr('MeanRadius'))
    """TBD."""

    credits = Unicode('').tag(xml=XmlSer.text_elem('Credits'))
    """Textual credits for the image originator."""

    credits_url = Unicode('').tag(xml=XmlSer.text_elem('CreditsUrl'))
    """A URL giving the source of the image or more information about its creation."""

    thumbnail_url = Unicode('').tag(xml=XmlSer.text_elem('ThumbnailUrl'))
    """A URL to a standard WWT thumbnail representation of this imageset."""

    description = Unicode('').tag(xml=XmlSer.text_elem('Description'))
    """A textual description of the imagery."""

    msr_community_id = Int(0).tag(xml=XmlSer.attr('MSRCommunityId'))
    """The ID number of the WWT Community that this content came from."""

    msr_component_id = Int(0).tag(xml=XmlSer.attr('MSRComponentId'))
    """The ID number of this content item on the WWT Communities system."""

    permission = Int(0).tag(xml=XmlSer.attr('Permission'))
    "TBD."

    def _tag_name(self):
        return 'ImageSet'

    def mutate_urls(self, mutator):
        if self.url:
            self.url = mutator(self.url)

        if self.dem_url:
            self.dem_url = mutator(self.dem_url)

        if self.credits_url:
            self.credits_url = mutator(self.credits_url)

        if self.thumbnail_url:
            self.thumbnail_url = mutator(self.thumbnail_url)

    def set_position_from_wcs(self, headers, width, height, place=None, fov_factor=1.7):
        """Set the positional information associated with this imageset to match a set
        of WCS headers.

        Parameters
        ----------
        headers : :class:`~astropy.io.fits.Header` or string-keyed dict-like
          A set of FITS-like headers including WCS keywords such as ``CRVAL1``.
        width : positive integer
          The width of the image associated with the WCS, in pixels.
        height : positive integer
          The height of the image associated with the WCS, in pixels.
        place : optional :class:`~wwt_data_formats.place.Place`
          If specified, the centering and zoom level of the :class:`~wwt_data_formats.place.Place`
          object will be set to match the center and size of this image.
        fov_factor : optional float
          If *place* is provided, its zoom level will be set so that the
          angular height of the client viewport is this factor times the
          angular height of the image. The default is 1.7.

        Returns
        -------
        self
          For convenience in chaining function calls.

        Notes
        -----
        Certain of the ImageSet parameters take on different meanings depending on
        whether the image in question is a tiled "study" or not. This method will alter
        its behavior depending on whether the :attr:`tile_levels` attribute is greater
        than zero. If you are computing coordinates for a tiled study, make sure to set
        this parameter *before* calling this function.

        For the time being, the WCS must be equatorial using the gnomonic
        (``TAN``) projection.

        Required keywords in *headers* are:

        - ``CTYPE1`` and ``CTYPE2``
        - ``CRVAL1`` and ``CRVAL2``
        - ``CRPIX1`` and ``CRPIX2``
        - Either:
          - ``CDELT1``, ``CDELT2``, ``PC1_1``, and ``PC1_2``; or
          - ``CD1_1``, ``CD2_2``

        If present ``PC1_2``, ``PC2_1``, ``CD1_2``, and/or ``CD2_1`` are used.
        If absent, they are assumed to be zero.

        """
        if headers['CTYPE1'] != 'RA---TAN' or headers['CTYPE2'] != 'DEC--TAN':
            raise ValueError('WCS coordinates must be in an equatorial/TAN projection')

        # Figure out the stuff we need from the headers.

        ra_deg = headers['CRVAL1']
        dec_deg = headers['CRVAL2']
        crpix_x = headers['CRPIX1'] - 1
        crpix_y = headers['CRPIX2'] - 1

        if 'CD1_1' in headers:
            cd1_1 = headers['CD1_1']
            cd2_2 = headers['CD2_2']
            cd1_2 = headers.get('CD1_2', 0.0)
            cd2_1 = headers.get('CD2_1', 0.0)

            if cd1_1 * cd2_2 - cd1_2 * cd2_1 < 0:
                cd_sign = -1
            else:
                cd_sign = 1

            rot_rad = math.atan2(-cd_sign * cd1_2, cd2_2)
            scale_x = math.sqrt(cd1_1**2 + cd2_1**2) * cd_sign
            scale_y = math.sqrt(cd1_2**2 + cd2_2**2)
        else:
            scale_x = headers['CDELT1']
            scale_y = headers['CDELT2']
            pc1_1 = headers.get('PC1_1', 1.0)
            pc2_2 = headers.get('PC2_2', 1.0)
            pc1_2 = headers.get('PC1_2', 0.0)
            pc2_1 = headers.get('PC2_1', 0.0)

            det = pc1_1 * pc2_2 - pc1_2 * pc2_1

            if det < 0:
                pc_sign = -1
            else:
                pc_sign = 1

            rot_rad = math.atan2(-pc_sign * pc1_2, pc2_2)

            # I am not sure if this is "supposed" to be allowed, but I've seen it.
            rtdet = math.sqrt(pc_sign * det)
            scale_x *= rtdet
            scale_y *= rtdet

        # This is our best effort to make sure that the view centers on the
        # center of the image.

        try:
            from astropy.wcs import WCS
        except:
            center_ra_deg = ra_deg
            center_dec_deg = dec_deg
        else:
            wcs = WCS(headers)
            center = wcs.pixel_to_world(height / 2, width / 2)
            center_ra_deg = center.ra.deg
            center_dec_deg = center.dec.deg

        # Now, assign the fields

        self.data_set_type = DataSetType.SKY
        self.width_factor = 2
        self.center_x = ra_deg
        self.center_y = dec_deg
        self.rotation_deg = rot_rad * 180 / math.pi

        if self.tile_levels > 0:  # are we tiled?
            self.projection = ProjectionType.TAN
            self.offset_x = (width / 2 - crpix_x) * abs(scale_x)
            self.offset_y = (height / 2 - crpix_y) * scale_y
            self.base_degrees_per_tile = scale_y * 256 * 2**self.tile_levels
        else:
            self.projection = ProjectionType.SKY_IMAGE
            self.offset_x = crpix_x
            self.offset_y = crpix_y
            self.base_degrees_per_tile = scale_y

        if place is not None:
            place.data_set_type = DataSetType.SKY
            place.rotation_deg = 0.  # I think this is better than propagating the image rotation?
            place.ra_hr = center_ra_deg / 15.
            place.dec_deg = center_dec_deg
            # It is hardcoded that in sky mode, zoom_level = height of client FOV * 6.
            place.zoom_level = height * scale_y * fov_factor * 6

        return self


    def wcs_headers_from_position(self):
        """Compute a set of WCS headers for this ImageSet's positional information.

        Returns
        -------
        A string-keyed dict-like containing FITS/WCS header keywords such as
        ``CTYPE1``, ``CRPIX1``, etc.

        Notes
        -----
        At the moment, this function only works for ImageSets with a
        projection type of ``SKY_IMAGE``. Support for other projections
        *might* be added later, if the need arises..

        """
        rv = {
            'CTYPE1': 'RA---TAN',
            'CTYPE2': 'DEC--TAN',
            'CRVAL1': self.center_x,
            'CRVAL2': self.center_y,
        }

        if self.projection != ProjectionType.SKY_IMAGE:
            raise NotImplementError('wcs_headers_from_position() only works if projection=SKY_IMAGE')

        rv['CRPIX1'] = self.offset_x + 1
        rv['CRPIX2'] = self.offset_y + 1
        rv['CDELT2'] = self.base_degrees_per_tile  # = scale_y, above
        rv['CDELT1'] = -self.base_degrees_per_tile  # AFAICT, non-square pixels can't be expressed

        c = math.cos(self.rotation_deg * math.pi / 180)
        s = math.sin(self.rotation_deg * math.pi / 180)

        rv['PC1_1'] = c
        rv['PC1_2'] = -s
        rv['PC2_1'] = s
        rv['PC2_2'] = c

        return rv



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the .NET Foundation
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

__all__ = '''
Folder
fetch_folder_tree
make_absolutizing_url_mutator
walk_cached_folder_tree
'''.split()

import os.path
import re
import requests
from traitlets import Bool, Instance, Int, List, Unicode, Union, UseEnum
from xml.etree import ElementTree as etree

from . import LockedXmlTraits, XmlSer
from .abcs import UrlContainer
from .enums import FolderType

class Folder(LockedXmlTraits, UrlContainer):
    """A grouping of WWT content assets.

    Children can be: places (aka "Items"), imagesets, linesets, tours,
    folders, or IThumbnail objects (to be explored).

    """
    name = Unicode('').tag(xml=XmlSer.attr('Name'))
    group = Unicode('Explorer').tag(xml=XmlSer.attr('Group'))
    url = Unicode('').tag(xml=XmlSer.attr('Url'))
    """The URL at which the full contents of this folder can be downloaded in WTML
    format.

    """
    thumbnail = Unicode('').tag(xml=XmlSer.attr('Thumbnail'))
    browseable = Bool(True).tag(xml=XmlSer.attr('Browseable'))
    searchable = Bool(True).tag(xml=XmlSer.attr('Searchable'))
    type = UseEnum(
        FolderType,
        default_value = FolderType.SKY,
    ).tag(xml=XmlSer.attr('Type'))
    sub_type = Unicode('').tag(xml=XmlSer.attr('SubType'))
    msr_community_id = Int(0).tag(xml=XmlSer.attr('MSRCommunityId'))
    """The ID number of the WWT Community that this content came from."""

    msr_component_id = Int(0).tag(xml=XmlSer.attr('MSRComponentId'))
    """The ID number of this content item on the WWT Communities system."""

    permission = Int(0).tag(xml=XmlSer.attr('Permission'))
    "TBD."

    children = List(
        trait = Union([
            Instance('wwt_data_formats.folder.Folder', args=()),
            Instance('wwt_data_formats.place.Place', args=()),
            Instance('wwt_data_formats.imageset.ImageSet', args=()),
        ]),
        default_value = ()
    ).tag(xml=XmlSer.inner_list())

    def _tag_name(self):
        return 'Folder'

    def walk(self, download=False):
        yield (0, (), self)

        for index, child in enumerate(self.children):
            if isinstance(child, Folder):
                if not len(child.children) and child.url and download:
                    url = child.url
                    child = Folder.from_url(url)
                    child.url = url
                    self.children[index] = child

                for depth, path, subchild in child.walk(download=download):
                    yield (depth + 1, (index,) + path, subchild)
            else:
                yield (1, (index,), child)

    def mutate_urls(self, mutator):
        if self.url:
            self.url = mutator(self.url)
        if self.thumbnail:
            self.thumbnail = mutator(self.thumbnail)

        for c in self.children:
            c.mutate_urls(mutator)


def make_absolutizing_url_mutator(baseurl):
    """Return a function that makes relative URLs absolute.

    Parameters
    ----------
    baseurl : string, absolute URL
        The absolute URL with which to combine relative URLs

    Returns
    -------
    A mutator function suitable for use with :meth:`wwt_data_formats.abcs.UrlContainer.mutate_urls`.

    Notes
    -----
    This function is designed for usage with :meth:`wwt_data_formats.abcs.UrlContainer.mutate_urls`.
    It returns a mutator function that can be passed to this method. The mutator will take
    relative URLs and make them absolute by combining them with the *baseurl* argument. Input URLs
    that are already absolute will be unchanged.

    """
    from urllib.parse import urljoin, urlsplit

    def mutator(url):
        if not url:
            return url
        if urlsplit(url).netloc:
            return url  # this URL is absolute
        return urljoin(baseurl, url)

    return mutator


def _sanitize_name(name):
    s = re.sub('[^-_a-zA-Z0-9]+', '_', name)
    s = re.sub('^_+', '', s)
    s = re.sub('_+$', '', s)
    return s


def fetch_folder_tree(root_url, root_cache_path, on_fetch=None):
    done_urls = set()

    def get_folder(url):
        if url in done_urls:
            return None, None

        on_fetch(url)
        resp = requests.get(url)
        resp.encoding = 'utf-8-sig'  # see LockedXmlTraits.from_urL()
        elem = etree.fromstring(resp.text)
        done_urls.add(url)
        return resp.text, Folder.from_xml(elem)

    root_text, root_folder = get_folder(root_url)
    with open(os.path.join(root_cache_path, 'index.wtml'), 'wt') as f:
        f.write(root_text)

    def walk(cur_folder, cur_cache_path):
        for index, child in enumerate(cur_folder.children):
            if not isinstance(child, Folder):
                continue

            text = None
            subdir_base = f'{index:03d}_{_sanitize_name(child.name)}'
            child_cache_path = os.path.join(cur_cache_path, subdir_base)

            if not len(child.children) and child.url:
                text, child = get_folder(child.url)
                if child is None:
                    continue

                os.makedirs(child_cache_path, exist_ok=True)
                with open(os.path.join(child_cache_path, 'index.wtml'), 'wt') as f:
                    f.write(text)

            walk(child, child_cache_path)

    walk(root_folder, root_cache_path)


def walk_cached_folder_tree(root_cache_path):
    seen_urls = set()

    root_folder = Folder.from_file(os.path.join(root_cache_path, 'index.wtml'))

    def walk(cur_treepath, cur_folder, cur_cache_path):
        yield (cur_treepath, cur_folder)

        for index, child in enumerate(cur_folder.children):
            child_treepath = cur_treepath + (index,)

            if not isinstance(child, Folder):
                yield (child_treepath, child)
            else:
                subdir_base = f'{index:03d}_{_sanitize_name(child.name)}'
                child_cache_path = os.path.join(cur_cache_path, subdir_base)

                if not len(child.children) and child.url:
                    if child.url in seen_urls:
                        continue

                    seen_urls.add(child.url)
                    child = Folder.from_file(os.path.join(child_cache_path, 'index.wtml'))

                for sub_treepath, sub_child in walk(child_treepath, child, child_cache_path):
                    yield (sub_treepath, sub_child)

    for info in walk((), root_folder, root_cache_path):
        yield info
